'use strict';
const _ = require('lodash');
const presets = require('../presets');
const fs = require('fs');
const path = require('path');
const DEFAULTS = {
    path: {
        type: 'string',
        default: process.cwd()
    },
    changelogFile: {
        type: 'string',
        default: 'CHANGELOG.md'
    },
    historyFile: {
        type: 'string',
        default: path.join('.versionbot', 'CHANGELOG.yml')
    },
    defaultInitialVersion: {
        type: 'string',
        default: '0.0.1'
    },
    gitDirectory: {
        type: 'string',
        default: '.git'
    },
    parseFooterTags: {
        type: 'boolean',
        default: true
    },
    lowerCaseFooterTags: {
        type: 'boolean',
        default: true
    },
    editChangelog: {
        type: 'boolean',
        default: true
    },
    editVersion: {
        type: 'boolean',
        default: true
    },
    subjectParser: {
        type: 'function',
        default: _.identity,
        allowsPresets: true
    },
    bodyParser: {
        type: 'function',
        default: _.identity,
        allowsPresets: true
    },
    includeCommitWhen: {
        type: 'function',
        default: 'has-changetype',
        allowsPresets: true
    },
    transformTemplateData: {
        type: 'function',
        default: 'changelog-entry',
        allowsPresets: true
    },
    transformTemplateDataAsync: {
        type: 'function',
        default: (data, cb) => {
            return cb(null, data);
        },
        allowsPresets: true
    },
    includeMergeCommits: {
        type: 'boolean',
        default: false
    },
    getChangelogDocumentedVersions: {
        type: 'function',
        default: 'changelog-headers',
        allowsPresets: true
    },
    getCurrentBaseVersion: {
        type: 'function',
        default: 'latest-documented',
        allowsPresets: true
    },
    getIncrementLevelFromCommit: {
        type: 'function',
        default: 'change-type-or-subject',
        allowsPresets: true
    },
    incrementVersion: {
        type: 'function',
        default: 'semver',
        allowsPresets: true
    },
    getGitReferenceFromVersion: {
        type: 'function',
        default: 'v-prefix',
        allowsPresets: true
    },
    addEntryToChangelog: {
        type: 'function',
        default: {
            preset: 'prepend',
            fromLine: 6
        },
        allowsPresets: true
    },
    updateVersion: {
        type: ['function', 'array'],
        default: 'npm',
        allowsPresets: true
    },
    template: {
        type: 'string',
        default: 'default',
        allowsPresets: true
    },
    addEntryToHistoryFile: {
        type: 'function',
        default: 'yml-prepend',
        allowsPresets: true
    }
};
exports.isPresetProperty = (presetsHash, propertyName, value) => {
    if (_.isPlainObject(value) && _.isString(value.preset)) {
        return Boolean(presetsHash[propertyName][value.preset]);
    }
    if (_.isString(value)) {
        return Boolean(presetsHash[propertyName][value]);
    }
    return false;
};
exports.parsePresetDefinition = (value) => {
    if (_.isString(value)) {
        return {
            name: value,
            options: {}
        };
    }
    return {
        name: value.preset,
        options: _.omit(value, 'preset')
    };
};
exports.getPropertyParsedValue = (propertyName, propertyDescription, data) => {
    const currentValue = _.get(data, propertyName);
    if (_.isUndefined(currentValue) || _.isNull(currentValue)) {
        return propertyDescription.default;
    }
    return currentValue;
};
exports.isPropertyValueValid = (propertyDescription, value) => {
    const typeCheck = (v, d) => {
        if (d === 'array') {
            return _.isArray(v);
        }
        if (d === 'object') {
            return _.isObject(v);
        }
        return typeof v === d;
    };
    if (_.isArray(value) && _.isArray(propertyDescription.type)) {
        return _.every(value, (v) => {
            return _.some(propertyDescription.type, (d) => {
                return typeCheck(v, d);
            });
        });
    }
    if (_.isArray(propertyDescription.type)) {
        return _.some(propertyDescription.type, (d) => {
            return typeCheck(value, d);
        });
    }
    if (_.isArray(value)) {
        return _.every(value, (v) => {
            return typeCheck(v, propertyDescription.type);
        });
    }
    return typeCheck(value, propertyDescription.type);
};
exports.parsePreset = (propertyDescription, presetsHash, propertyName, value) => {
    const propertyPresets = _.get(presetsHash, propertyName, {});
    const presetDefinition = exports.parsePresetDefinition(value);
    const matchedPreset = _.get(propertyPresets, presetDefinition.name);
    if (!matchedPreset) {
        throw new Error(`Invalid preset: ${propertyName} -> ${presetDefinition.name}`);
    }
    if (propertyDescription.type === 'function' || _.find(propertyDescription.type, (t) => {
        return t === 'function';
    })) {
        return _.partial(matchedPreset, presetDefinition.options);
    }
    return matchedPreset;
};
exports.parse = (data) => {
    return _.mapValues(DEFAULTS, (propertyDescription, propertyName) => {
        let value = exports.getPropertyParsedValue(propertyName, propertyDescription, data);
        if (propertyDescription.allowsPresets && exports.isPresetProperty(presets, propertyName, value)) {
            return exports.parsePreset(propertyDescription, presets, propertyName, value);
        }
        if (_.isArray(value)) {
            value = value.map((v) => {
                if (propertyDescription.allowsPresets && exports.isPresetProperty(presets, propertyName, v)) {
                    return exports.parsePreset(propertyDescription, presets, propertyName, v);
                }
                return v;
            });
        }
        if (!exports.isPropertyValueValid(propertyDescription, value)) {
            throw new Error([
                `Invalid option value: ${value}.`,
                `The \`${propertyName}\` option expects a ${_.isArray(propertyDescription.type)
                    ? propertyDescription.type.join(' or ') : propertyDescription.type},`,
                `but instead got a ${typeof value}.`
            ].join(' '));
        }
        return value;
    });
};
exports.load = (file) => {
    try {
        return require(file);
    }
    catch (error) {
        if (error instanceof SyntaxError) {
            throw new Error(`Syntax error in configuration file: ${file}`);
        }
        throw error;
    }
};
exports.hasDefaultConfigFile = (defaultPath) => {
    if (fs.existsSync(defaultPath)) {
        return defaultPath;
    }
    return '__NO_CONFIG';
};
//# sourceMappingURL=configuration.js.map