"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const yaml = require("js-yaml");
const _ = require("lodash");
const tags = require("./tags");
exports.GIT_LOG_YAML_PRETTY_FORMAT = [
    '- hash: %H',
    '  subject: >-',
    '    %s',
    '  body: |-',
    '    XXX',
    '    %w(0,0,4)%b',
].join('%n');
exports.getGitLogRevisionRange = ({ startReference, endReference = 'HEAD', } = {}) => {
    if (startReference) {
        return `${startReference}..${endReference}`;
    }
    return endReference;
};
exports.getGitLogCommandArgumentsThatOutputYaml = (options) => {
    if (!options) {
        throw new Error('Missing the options');
    }
    if (!options.gitDirectory) {
        throw new Error('Missing the gitDirectory option');
    }
    const command = [
        `--git-dir=${options.gitDirectory}`,
        'log',
        `--pretty=${exports.GIT_LOG_YAML_PRETTY_FORMAT}`,
    ];
    if (!options.includeMergeCommits) {
        command.push('--no-merges');
    }
    command.push(exports.getGitLogRevisionRange(options));
    return command;
};
exports.parseGitLogYAMLOutput = (output, { subjectParser = _.identity, bodyParser = _.identity, parseFooterTags = true, lowerCaseFooterTags = false, } = {}) => {
    return _.map(yaml.safeLoad(output), (commit) => {
        if (_.isUndefined(commit.subject)) {
            throw new Error('Invalid commit: no subject');
        }
        if (_.isUndefined(commit.body)) {
            throw new Error('Invalid commit: no body');
        }
        commit.body = _.chain(commit.body)
            .split('\n')
            .tail()
            .join('\n')
            .value();
        const commitObj = {
            subject: subjectParser(commit.subject),
            hash: commit.hash,
        };
        if (!parseFooterTags) {
            return Object.assign({}, commitObj, { body: bodyParser(commit.body) });
        }
        const commitDescription = _.chain(commit.body)
            .split('\n')
            .reduceRight((accumulator, commitLine) => {
            const isTag = tags.isTagLine(commitLine);
            if (_.isEmpty(commitLine.trim()) || !isTag) {
                accumulator.considerTags = false;
            }
            if (accumulator.considerTags && isTag) {
                accumulator.footer.unshift(commitLine);
            }
            else {
                accumulator.body.unshift(commitLine);
            }
            return accumulator;
        }, {
            considerTags: true,
            body: [],
            footer: [],
        })
            .value();
        return Object.assign({}, commitObj, { body: bodyParser(commitDescription.body.join('\n')), footer: tags.parseFooterTagLines(commitDescription.footer, {
                lowerCaseFooterTags,
            }) });
    });
};
//# sourceMappingURL=git-log.js.map