"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const childProcess = require("child_process");
const _ = require("lodash");
const gitLog = require("./git-log");
const semver = require("./semver");
const template = require("./template");
exports.readCommitHistory = (gitDirectory, options, callback) => {
    if (!gitDirectory) {
        throw new Error('Missing gitDirectory');
    }
    const command = gitLog.getGitLogCommandArgumentsThatOutputYaml({
        gitDirectory,
        startReference: options.startReference,
        endReference: options.endReference,
        includeMergeCommits: options.includeMergeCommits,
    });
    const child = childProcess.spawn('git', command);
    let stdout = '';
    child.stdout.on('data', data => {
        stdout += data;
    });
    child.stderr.on('data', data => {
        child.kill();
        return callback(new Error(data.toString()));
    });
    child.on('error', error => {
        child.kill();
        return callback(error);
    });
    child.on('close', code => {
        if (code !== 0) {
            return callback(new Error(`Child process exitted with error code: ${code}`));
        }
        const parsedCommits = gitLog.parseGitLogYAMLOutput(stdout, {
            subjectParser: options.subjectParser,
            bodyParser: options.bodyParser,
            parseFooterTags: options.parseFooterTags,
            lowerCaseFooterTags: options.lowerCaseFooterTags,
        });
        return callback(undefined, parsedCommits);
    });
};
exports.calculateNextVersion = (commits, options) => {
    if (!options) {
        throw new Error('Missing the options argument');
    }
    if (!options.currentVersion) {
        throw new Error('Missing the currentVersion option');
    }
    if (!options.getIncrementLevelFromCommit) {
        throw new Error('Missing the getIncrementLevelFromCommit option');
    }
    if (!_.isFunction(options.getIncrementLevelFromCommit)) {
        throw new Error(`Invalid getIncrementLevelFromCommit option: ${options.getIncrementLevelFromCommit}`);
    }
    if (!options.incrementVersion) {
        throw new Error('Missing the incrementVersion option');
    }
    if (!_.isFunction(options.incrementVersion)) {
        throw new Error(`Invalid incrementVersion option: ${options.incrementVersion}`);
    }
    const incrementLevel = semver.calculateNextIncrementLevel(commits, {
        getIncrementLevelFromCommit: options.getIncrementLevelFromCommit,
    });
    if (!incrementLevel) {
        return options.currentVersion;
    }
    return options.incrementVersion(options.currentVersion, incrementLevel);
};
exports.generateChangelog = (commits, options, callback) => {
    if (_.isEmpty(commits)) {
        throw new Error('No commits to generate the CHANGELOG from');
    }
    if (!options) {
        throw new Error('Missing the options argument');
    }
    if (!options.template) {
        throw new Error('Missing the template option');
    }
    if (!options.version) {
        throw new Error('Missing the version option');
    }
    const { date = new Date(), includeCommitWhen = _.constant(true), transformTemplateData = _.identity, transformTemplateDataAsync = (data, cb) => {
        return cb(undefined, data);
    }, } = options;
    if (!(date instanceof Date)) {
        throw new Error(`Invalid date option: ${date}`);
    }
    semver.checkValid(options.version);
    const templateData = transformTemplateData({
        commits: _.filter(commits, includeCommitWhen),
        version: options.version,
        date,
    });
    transformTemplateDataAsync(templateData, (error, transformedTemplate) => {
        if (error) {
            return callback(error);
        }
        return callback(undefined, template.render(options.template, transformedTemplate), transformedTemplate);
    });
};
//# sourceMappingURL=versionist.js.map